// FreePaTris - A Tetris clone written in FreePascal
// By Marcus Riemer (Inf8803)
// 
// I tried using a gamestate'ish mechanism, but due too lack of time (and knowledge)
// I decided to use a rather simple version:
//      Every gamestate has its own unit and should define:
//               DoGamestate:    Takes the time since the last frame and returns the id of the
//                               gamestate that shal be shown the next frame.
//               EnterGamestate: Called whenever the gamestate is "freshly" entered
// Because I didn't have the time to work out the way OOP works in FreePascal (especially
// regarding inheritance and virtual methods) I decided to hard-code the gamestate in the
// "Run" procedure of the program.
//
// Compiler notice:
// Should be compiled with ObjectPascal extensions on. Otherwise things like
// exceptions and C Style enumerations will cause compilation errors.
// 
// 22.10.2009
//  - Creation
// 24.10.2009
//  - General drawing and collision
// 25.10.2009
//  - Rotation
//  - Documentation

program FreePaTris;

uses
	Types, // Common used types
	Menu, // The Menu gamestate
	Game, // The Tetris gamestate
	SysUtils, // Timing
	Crt; // Console operations
	
// Use this to make sure the correct StateEnter and StateExit procedures
// are called on gamestate changes
// Added parameter ingame in a hurry, because unit global variables
// aren't allowed
//
// Parameters:
//  current: the ID of the current gamestate
//  next: the id of the coming gamestate
//  ingame: the game gamestate container, which may be altered by EnterGame()
//
// Calls
//  - EnterGame
function ChangeGamestate(current : TGamestateId; next : TGamestateId;
						 var ingame : TGameState) : TGamestateId;
begin
	// Calling the correct StateExit procedure

	// Calling the correct StateEnter procedure
	case next of
		StateGame: ingame := EnterGame();
	end;
	
	ChangeGamestate := next;
end;

// Main game loop
// - Clears the screen at the begin of each frame
// - Manages gamestates
// - Manages time
//
// Calls
//  - DoMenu
//  - DoGame
//  - ChangeGamestate
procedure Run();
var
	_timeThisFrame : TTimeStamp; // Time passed since program started in this frame
	_timeLastFrame : TTimeStamp; // Time passed since the program started in the last frame
	_timeFrame : comp; // The frametime for the current frame in ms
	_currentGamestate : TGamestateId; // The id of the current gamestate
	_nextGamestate: TGamestateId; // The id of the next gamestate

	// As not even unit-global variables are allowed,
	// this record will be passed around
	_gamestate : TGameState;

begin
	// Default assignments
	_timeThisFrame := DateTimeToTimeStamp(Now());
	_timeLastFrame := DateTimeToTimeStamp(Now());
	_timeFrame := 0;
	_currentGamestate := StateMenu;
	_nextGamestate := StateQuit;

	// Run the game, as long as the gamestate "permits" it
	while (_currentGamestate <> StateQuit) do
	begin
		// Doing the timing
		// - "Stop" the current time
		// - Subtract the current time from the time in the last frame to get the frametime in ms
		_timeThisFrame := DateTimeToTimeStamp(Now());
		_timeFrame := TimeStampToMSecs(_timeThisFrame) - TimeStampToMSecs(_timeLastFrame);

		// Calling the correct DoGamestate function and setting
		// the return value as the state for the next frame
		case _currentGamestate of
			StateMenu: _nextGamestate := DoMenu(_timeFrame);
			StateGame: _nextGamestate := DoGame(_timeFrame, _gamestate);
		end;
		
		// Do we need to enter another gamestate?
		if _nextGamestate <> _currentGamestate then begin
			// Yup, so we need to make sure the appropriate EnterState and ExitState
			// procedures are called
			ChangeGamestate(_currentGamestate, _nextGamestate, _gamestate);
			
			_currentGamestate := _nextGamestate;
		end;
		
		// Copy over time  of this frame for next frame
		_timeLastFrame := _timeThisFrame;
	end;
				
end;

// Program start
//
// Calls
//  - Run
begin
	// Initialise random number generator
	Randomize();

	// Turn of the cursor
	CursorOff();

	// Clear the entire screen
	clrscr();
 
    // Sort of center the window
    Window(WindowX, WindowY, WindowX + 40, WindowX + 40);
    
	// Run the game
	Run();
end.
